﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Input;
using Windows.ApplicationModel.Appointments;
using Windows.ApplicationModel.VoiceCommands;

namespace TheCalculator.ViewModels {
    public class CalculatorViewModel: INotifyPropertyChanged {
        private decimal _calcCache;
        private string _opration;
        private const string _ADD = "Add";
        private const string _SUBTRACT = "Subtract";
        private const string _MULTIPLY = "Multiply";
        private const string _DIVIDE = "Divide";
        private const string _cannotDivideByZero = "No Divide By Zero!";
        private bool _clearCalcTxt = false;

        private string _calcText;
        public string CalcText { 
            get { return _calcText; } 
            set {
                if (_clearCalcTxt) {
                    _calcText = string.Empty;
                    _clearCalcTxt = false;
                }
                _calcText = value;
                OnPropertyChanged("CalcText");
            }
        }

        private ICommand _commandButtonNumber;
        public ICommand CommandButtonNumber {
            get { 
                return _commandButtonNumber ?? (_commandButtonNumber = new Updater<string>((x) => { UpdateCalcText(x); })); 
            }
            set {
                _commandButtonNumber = value;
            }
        }

        private ICommand _commandButtonCalc;
        public ICommand CommandButtonCalc {
            get {
                return _commandButtonCalc ?? (_commandButtonCalc = new Updater<string>((x) => { CalcCommand(x); }));
            }
            set {
                _commandButtonCalc = value;
            }
        }

        private void ClearDivideByZero() {
            if (!string.IsNullOrEmpty(CalcText) && CalcText.Equals(_cannotDivideByZero))
                ClearCalcText();
        }

        public void CalcCommand(string calcType) {
            ClearDivideByZero();

            switch (calcType) {
                case "Clear": {
                        ClearCalcText();
                        ClearCalcCache();
                        break;
                    }
                case "Add": {
                        SetAddition();
                        break;
                    }
                case "Subtract": {
                        SetSubtraction();
                        break;
                    }
                case "Multiply": {
                        SetMultiplication();
                        break;
                    }
                case "Divide": {
                        SetDivision();
                        break;
                    }
                case "Equals": {
                        ComputeEquals();
                        break;
                    }
            }
        }

        public void UpdateCalcText(string text) {
            ClearDivideByZero();

            CalcText += text;
        }

        private void ClearCalcText() {
            CalcText = string.Empty;
        }

        private void ComputeEquals() {
            if (string.IsNullOrEmpty(CalcText) || string.IsNullOrEmpty(_opration))
                return;

            switch (_opration) {
                case _ADD:  {
                        ComputeAddition();
                        break;
                    }
                case _SUBTRACT: {
                        ComputeSubtraction();
                        break;
                    }
                case _MULTIPLY: {
                        ComputeMultiplication();
                        break;
                    }
                case _DIVIDE: {
                        ComputeDivision();
                        break;
                    }
            }

            if (CalcText.Equals(_cannotDivideByZero))
                return;

            _calcCache = Math.Abs(_calcCache);
            CalcText = _calcCache.ToString();
        }

        private void SetAddition() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = decimal.Parse(CalcText);
            _opration = _ADD;
            ClearCalcText();
        }

        private void ComputeAddition() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = _calcCache + decimal.Parse(CalcText);
        }

        private void SetSubtraction() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = decimal.Parse(CalcText);
            _opration = _SUBTRACT;
            ClearCalcText();
        }

        private void ComputeSubtraction() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = decimal.Parse(CalcText) - _calcCache;
        }

        private void SetMultiplication() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = decimal.Parse(CalcText);
            _opration = _MULTIPLY;
            ClearCalcText();
        }

        private void ComputeMultiplication() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = decimal.Parse(CalcText) * _calcCache;
        }

        private void SetDivision() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            _calcCache = decimal.Parse(CalcText);
            _opration = _DIVIDE;
            ClearCalcText();
        }

        private void ComputeDivision() {
            if (string.IsNullOrEmpty(CalcText))
                return;

            if (CalcText.Equals("0")) {
                CalcText = _cannotDivideByZero;
                _clearCalcTxt = true;
                return;
            }

            _calcCache = _calcCache / decimal.Parse(CalcText);
        }

        private void ClearCalcCache() {
            _calcCache = 0;
            _opration = default(string);
        }

        // boiler-plate
        public event PropertyChangedEventHandler PropertyChanged;
        protected virtual void OnPropertyChanged(string propertyName) {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        //protected bool SetField<T>(ref T field, T value, string propertyName) {
        //    if (EqualityComparer<T>.Default.Equals(field, value)) return false;
        //    field = value;
        //    OnPropertyChanged(propertyName);
        //    return true;
        //}

        public class Updater<T> : ICommand {
            private Action<T> _action;
            public event EventHandler CanExecuteChanged;

            public Updater(Action<T> action) {
                _action = action;
            }

            public bool CanExecute(object parameter) {
                return true;
            }

            public void Execute(object parameter) {
                _action?.Invoke((T)parameter);
            }
        }
    }
}
